/*
 * Citadel/UX Network Processor v1.52
 * designed and written by Art Cancro @ Uncensored Communications Group
 * see copyright.doc for copyright information
 * 
 *
 * History:
 * 08/01/88 AJC - Wrote v1.0
 *   /  /88 AJC - Rewrote inbound processor
 *              - Added shell command to do the dirty work of looking in
 *                the spoolin directory for inbound network files
 *   /  /88 AJC - Moved use table check into a function call
 *              - Changed usetable code to use originating message ID ("I"
 *                field) in table rather than message time. Time is still used
 *                if no I field is present.
 *   /  /88 AJC - Modified outbound side to not send a system its own messages
 * 10/28/88 AJC - Fixed bug that would screw up the use table if the
 *                originating system's name exceeded 19 characters.
 *              - Increased size of minfo buffers to handle big Henge names.
 *              - Annoying error messages from empty spoolin diverted to null
 * 11/25/88 AJC - Added software to look at P (path) field and not send a
 *		  message to a system if it's already been there
 * 05/30/89 AJC - Added mail routing based on D (destination system) field
 *              - Added routines to accept binary mail
 * 06/11/89 AJC - Oops! Fixed bug that caused mail arriving in rooms other than
 *                mail to be posted publicly AND privately. (One Nasty Bug)
 * 08/15/89 AJC - Read entire use table into memory during inbound processing,
 *                which writes it back to disk when finished. NOTE! If you have
 *                multiple netproc processes running concurrently, you MUST add
 *                some sort of locking code!
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include "citadel.h"

long atol();
long lseek();

struct recentmsg usemsgs[PREVMSGS];
int usepos;

struct minfo {
	char A[30];
	long I;
	char N[30];
	char O[30];
	char R[30];
	long T;
	char D[30];
	};


long finduser(file,name)
int file;
char name[];
{
	struct usersupp temp;
	int a;
	long bottom,top,guess,obottom,otop;

	obottom=(-1L); otop=(-1L);

	bottom=0L;
	top=(lseek(file,0L,2)/(long)sizeof(struct usersupp));

FPT:	if ((obottom==bottom)&&(otop==top)) {
		return(-1L); }
	obottom=bottom; otop=top;
	guess=(((top-bottom)/2)+bottom);
	lseek(file,(guess*(long)sizeof(struct usersupp)),0);
	read(file,&temp,sizeof(struct usersupp));
	a=strucmp(temp.fullname,name);
	if (a==0)
		
		return(lseek(file,(guess*(long)sizeof(struct usersupp)),0));
	if (a<0) {
		bottom=guess;
		goto FPT; }
	if (a>0) {
		top=guess;
		goto FPT; }
}



main(argc,argv)
int argc;
char *argv[];
{
	if (argc==2) outprocess(argv[1]);
	inprocess();
	exit(0);
}

inprocess() {	/* Process incoming files in ./network/spoolin */
	FILE *fp,*message;
	struct minfo minfo;
	struct recentmsg recentmsg;
	struct smreturn smreturn;
	char tname[50],aaa[100],iname[50];
	int a;

	sprintf(tname,"/tmp/net.t%d",getpid());	/* temp file name */
	sprintf(iname,"/tmp/net.i%d",getpid());	/* temp file name */

	if (readusetable()==1) {
		fprintf(stderr,"netproc: no use table online\n");
		fprintf(stderr,
	"         inbound is already running or netsetup needs to be run\n");
		return(1);
		}

	/* Let the shell do the dirty work. Get ALL data from spoolin */
sprintf(aaa,"for x in network/spoolin/*; do cat <$x; rm $x; done 2>/dev/null");
	fp=(FILE *)popen(aaa,"r");

NXMSG:	/* Break out the next message */
	do {
		a=getc(fp);
		} while((a!=255)&&(a>=0));
	if (a<0) goto ENDSTR;
	message=fopen(tname,"w");
	putc(255,message);
ZFLD:	do {
		a=getc(fp);
		putc(a,message);
		} while(a!=0);
	a=getc(fp);
	putc(a,message);
	if (a!='M') goto ZFLD;
	do {
		a=getc(fp);
		putc(a,message);
		} while(a>0);
	fclose(message);

	/* process the individual mesage */
	minfo.D[0]=0;
	msgfind(tname,&minfo);
	strncpy(recentmsg.RMnodename,minfo.N,19);
	recentmsg.RMnodename[19]=0;
	recentmsg.RMnum=minfo.I;
	a=usetable(&recentmsg);
	if (a==0) printf("Accept: ");
	if (a==1) printf("Reject: ");
	if (a==2) printf("Garbge: ");
	if (a>=3) printf("UKretn: ");
	printf("#%ld fm %s in %s @ %s\n",
		minfo.I,minfo.A,minfo.O,minfo.N);
	if (a!=0) goto SKMSG;
	if ((strucmp(minfo.D,NODENAME))&&(minfo.D[0]!=0)) { 
		sysinfo(minfo.D);
		ship_to(tname,minfo.D);
		}
	else {
		send_message(tname,&smreturn);
		if (minfo.R[0]!=0) {
			a=mail_to(minfo.R,&smreturn);
			if (a<0) file_message("Aide",&smreturn);
			}
		else {
			file_message(minfo.O,&smreturn);
			}
		}

SKMSG:	unlink(tname);
	goto NXMSG;

ENDSTR:	fclose(fp);
	unlink(iname);
	writeusetable();
	return(0);
}

readusetable() {
	int file;
	file=open("./network/prevmsgs",O_RDONLY);
	if (file<0) return(1);
	read(file,&usemsgs[0],(PREVMSGS*sizeof(struct recentmsg)));
	close(file);
	file=open("./network/prevmsgs.pos",O_RDONLY);
	usepos=0; read(file,&usepos,sizeof(usepos));
	close(file);
	return(0);
	}

writeusetable() {
	int file;
	file=open("./network/prevmsgs",O_WRONLY);
	write(file,&usemsgs[0],(PREVMSGS*sizeof(struct recentmsg)));
	close(file);
	file=open("./network/prevmsgs.pos",O_WRONLY);
	write(file,&usepos,sizeof(usepos));
	close(file);
	return(0);
	}

usetable(msg)		/* check use table	   */
struct recentmsg *msg;	/* returns 0 if ok to post */
	{
	int file3,a;
	struct recentmsg recent2;

	if ((msg->RMnum)<=0L) return(2);	/* can't use garbage ID's */
	for (a=0; a<PREVMSGS; ++a) {
		if ((!strucmp(msg->RMnodename,usemsgs[a].RMnodename))
			&&(msg->RMnum==usemsgs[a].RMnum)) {
				return(1);
				}
		}
	++usepos; if (usepos==PREVMSGS) usepos=0;
	usemsgs[usepos].RMnum=(msg->RMnum);
	strcpy(usemsgs[usepos].RMnodename,(msg->RMnodename));
	return(0);
}

msgfind(msgfile,buffer)
char *msgfile;
struct minfo *buffer;
	{
	struct tm *tm;
	int a,b,c,d,e,real,old,mtype,aflag;
	char aaa[100],bbb[100],ttm[3];
	FILE *fp;
		
	fp=fopen(msgfile,"rb");
	if (fp==NULL) {
			fprintf(stderr,"Can't open message file\n");
			return(1); }
	e=getc(fp);
	if (e!=255) {
		fprintf(stderr,"Incorrect message format\n");
		goto END; }
	mtype=getc(fp); aflag=getc(fp);
	buffer->I=0L;
	buffer->R[0]=0;

BONFGM:	b=getc(fp); if (b<0) goto END;
	if (b=='M') goto END;
	fpgetfield(fp,bbb);
	if (b=='A') strcpy(buffer->A,bbb);
	if (b=='O') strcpy(buffer->O,bbb);
	if (b=='N') strcpy(buffer->N,bbb);
	if (b=='R') strcpy(buffer->R,bbb);
	if (b=='D') strcpy(buffer->D,bbb);
	if (b=='T') buffer->T=atol(bbb);
	if (b=='I') buffer->I=atol(bbb);
	goto BONFGM;

END:	if (buffer->I==0L) buffer->I=buffer->T;
	fclose(fp);
	return(0);
}

fpgetfield(fp,string)
FILE *fp;
char string[];
{
int a,b,c;
strcpy(string,"");
a=0;
GFXYZAF:
	b=getc(fp);
	if (b<1) { string[a]=0; return(0); }
	string[a]=b;
	++a;
	if (b!=0) goto GFXYZAF;
}

send_message(filename,retbuf)	/* send a message to the master file */
char filename[];		/* tempfilename of proper message */
struct smreturn *retbuf;	/* return information */
{
int file,a,b,c,d,e,file2;
long aa,bb,cc,dd,ee,templen,hibytes,origpos,ffsource;
char aaa[100];
struct msgmain msgmain;

do {
	file=open("./MMstructure",O_RDWR);
	a=read(file,&msgmain,sizeof(struct msgmain));
	close(file);
		} while ( (msgmain.MMflags & MM_BUSY) !=0 );

	
	file=open("./MMstructure",O_RDWR);
	msgmain.MMflags=(msgmain.MMflags|MM_BUSY);
	a=write(file,&msgmain,sizeof(struct msgmain));
	close(file);
	origpos=msgmain.MMcurpos;

/* measure the message and count FF bytes */
ffsource=0L;
file=open(filename,O_RDONLY);
templen=0L;
do {
	a=0; b=read(file,&a,1);
	if (a==255) ++ffsource;
	templen=templen+(long)b;
	} while (b!=0);
close(file);

/* check for FF bytes */
hibytes=0L;
file=open("./msgmain",O_RDWR);
bb=lseek(file,msgmain.MMcurpos,0);

cc=lseek(file,0L,1);
for (bb=0L; bb<templen; ++bb) {
	if (cc>=MM_FILELEN) cc=lseek(file,0L,0);
	c=0; read(file,&c,1); ++cc;
	if (c>127) ++hibytes;		/* bump count if hi bit set */
	}

msgmain.MMlowest=msgmain.MMlowest+hibytes;
msgmain.MMhighest=msgmain.MMhighest+ffsource;

cc=lseek(file,msgmain.MMcurpos,0);
file2=open(filename,O_RDONLY);
cc=lseek(file,0L,1);
for (bb=0L; bb<templen; ++bb) {
	if (((long)cc)>=((long)MM_FILELEN)) cc=lseek(file,0L,0);
	read(file2,&b,1); 
	write(file,&b,1); ++cc;
	}
msgmain.MMcurpos=lseek(file,0L,1);
close(file2);	/* done with temp file */
close(file);	/* done with master file */

/* now update the message structure */
msgmain.MMflags=msgmain.MMflags & ~MM_BUSY;
file=open("./MMstructure",O_RDWR);
a=write(file,&msgmain,sizeof(struct msgmain));
close(file);

retbuf->smnumber=msgmain.MMhighest;
retbuf->smpos=origpos;
return(0);
}

strucmp(st1,st2)
char *st1;
char *st2;
{
char aaa[100],bbb[100];
int a;
	strcpy(aaa,st1); strcpy(bbb,st2);
	for (a=0; a<strlen(aaa); ++a) aaa[a]=tolower(aaa[a]);
	for (a=0; a<strlen(bbb); ++a) bbb[a]=tolower(bbb[a]);
	a=strcmp(aaa,bbb);
	return(a);
}

getline(buffer,file)
char buffer[];
int file; {
	int a,b,c;
	c=0; b=0;
	do {
		a=read(file,&b,1);
		if (b==10) b=0;
		if (a<1) b=0;
		buffer[c]=b; ++c; } while(b!=0);
	return(c-1);
}

pull_message(spos,file2) /* Pull a message out of the master file */
long spos; /* position in master file */
int file2; /* file to write to */
{
	int a,b,c;
	int file;
	long aa,bb,cc;
	char field[512];
	char fwork[512];

	file=open("msgmain",O_RDONLY);
	lseek(file,spos,0);

A:	for (a=0; a<3; ++a) {
		read(file,&b,1);
		write(file2,&b,1);
		}
PI:	b=0; read(file,&b,1); write(file2,&b,1);
	if (b=='M') goto B;
	getline(field,file);
	fwork[0]=0;
	if (b=='P') {
		strcpy(fwork,NODENAME);
		strcat(fwork,"!");
		}
	strcat(fwork,field);
	write(file2,fwork,strlen(fwork)+1);
	goto PI;

B:	cc=lseek(file,0L,1);
C:	if (cc>=MM_FILELEN) cc=lseek(file,0L,0);
	c=0; read(file,&c,1); ++cc;
	write(file2,&c,1);
	if (c!=0) goto C;

	close(file);
	return(0);
}
	
outprocess(sysname) /* send new room messages to sysname */
char *sysname;
{
	int a,b,c,roomnum,file,file2,file3,file4;
	long aa,bb,cc,lastseen;
	char aaa[300],bbb[300],spflname[100],tfname[50],sysflname[100];
	char roomname[50];
	char xroom[50],xsys[50],xpath[512];
	struct quickroom quickroom;
	struct fullroom fullroom;
	
	sprintf(sysflname,"network/systems/%s",sysname);
	file=open(sysflname,O_RDONLY);
	if (file<0) {
		fprintf("netproc: no system %s\n",sysname);
		return(1); }
	close(file);
	a=getpid();
	sprintf(spflname,"/tmp/%s.%d",NODENAME,a);
	file=creat(spflname,0666);
	chmod(spflname,0666);
	sprintf(tfname,"/tmp/nt%d",a);
	file2=creat(tfname,0666);
	chmod(tfname,0666);
	file3=open(sysflname,O_RDONLY);

	getline(aaa,file3);
	sprintf(bbb,"%s\n",aaa);	/* copy uucp command to temp file */
	write(file2,bbb,strlen(bbb));

NXRM:	/* get the next room off the list */
	a=getline(roomname,file3);
	if (a==0) goto FINISHED;
	sprintf(bbb,"%s\n",roomname);	/* copy it to the temp file */
	write(file2,bbb,strlen(bbb));
	getline(aaa,file3);
	lastseen=atol(aaa);

	/* look for the room in the local file */
	roomnum=(-1);
	file4=open("quickroom",O_RDONLY);
	for (a=0; a<MAXROOMS; ++a) {
		read(file4,&quickroom,sizeof(struct quickroom));
		if (!strucmp(quickroom.QRname,roomname)) roomnum=a;
		}
	close(file4);
	if (roomnum==(-1)) {
		fprintf(stderr,"netproc: no local room %s\n",roomname);
		goto NXRM;
		}
	sprintf(aaa,"./rooms/fullrm%d",roomnum);
	file4=open(aaa,O_RDONLY);
	read(file4,&fullroom,sizeof(struct fullroom));
	close(file4);
	
	/* go thru fullroom list, sending all new messages */
	for (a=0; a<MSGSPERRM; ++a) {
if (fullroom.FRnum[a]!=0L) b=check_message(fullroom.FRpos[a],xroom,xsys,xpath);
	if ((fullroom.FRnum[a]>lastseen)
		&& (strucmp(xsys,sysname))
		&& (checkpath(xpath,sysname)==1) 
			) {
			pull_message(fullroom.FRpos[a],file);
			lastseen=fullroom.FRnum[a];
			}
		}
	sprintf(aaa,"%ld\n",lastseen);
	write(file2,aaa,strlen(aaa));
	goto NXRM;

FINISHED:	/* finished spooling new messages */
	close(file);
	close(file2);
	close(file3);

	/* copy temporary sys file into permanent sys file */
	file2=open(tfname,O_RDONLY);
	file3=creat(sysflname,0666);
	chmod(sysflname,0666);
	do {
		b=0; a=read(file2,&b,1);
		if (a>0) write(file3,&b,1);
		} while(a>0);
	close(file2);
	close(file3);
	unlink(tfname);
	ship_to(spflname,sysname);		/* send the outbound file */
	unlink(spflname);			/* and then delete it     */
	return(0);
}

ship_to(filenm,sysnm)	/* send spool file filenm to system sysnm */
char *filenm;
char *sysnm; {
	char sysflnm[100];
	char commbuf1[100];
	char commbuf2[100];
	int sysflfd;

	sprintf(sysflnm,"./network/systems/%s",sysnm);
	sysflfd=open(sysflnm,O_RDONLY);
	getline(commbuf1,sysflfd);
	close(sysflfd);
	sprintf(commbuf2,commbuf1,filenm);
	return(system(commbuf2));
	}

check_message(pos,xroom,xsys,xpath) /* get room, system, path */
long pos;
char xroom[];
char xsys[];
char xpath[]; {
	FILE *fp;
	int a;

	strcpy(xroom,"");
	strcpy(xsys,"");
	strcpy(xpath,"");
	fp=fopen("msgmain","r");
	if (fp==NULL) return(1);
	fseek(fp,pos,0);

CMCZ:	do {
		a=getc(fp);
		if (a<0) {
			fclose(fp);
			return(2);
			}
		} while(a!=0);
CMCN:	a=getc(fp);
	if (a=='M') {
			fclose(fp);
			return(0);
			}
	if (a=='O') fpgetfield(fp,xroom);
	if (a=='N') fpgetfield(fp,xsys);
	if (a=='P') fpgetfield(fp,xpath);
	if ((a=='O')||(a=='N')||(a=='P')) goto CMCN;
	goto CMCZ;
}

int mail_to(user,where)
char user[];
struct smreturn *where; {
	struct usersupp usersupp;
	long uspos;
	int file,a;

	file=open("usersupp",O_RDWR);
	if (file<0) return(-1);
	uspos=finduser(file,user);
	if (uspos<0L) {
		close(file);
		return(-1);
		}
	read(file,&usersupp,sizeof(struct usersupp));
	for (a=0; a<(MAILSLOTS-1); ++a) {
		usersupp.mailnum[a]=usersupp.mailnum[a+1];
		usersupp.mailpos[a]=usersupp.mailpos[a+1];
		}
	usersupp.mailnum[MAILSLOTS-1]=where->smnumber;
	usersupp.mailpos[MAILSLOTS-1]=where->smpos;
	lseek(file,uspos,0);
	write(file,&usersupp,sizeof(struct usersupp));
	close(file);
	return(0);
	}

file_message(room,where)
char room[];
struct smreturn *where; {
	struct quickroom quickroom;
	struct fullroom fullroom;
	int rmnum,a,file;
	char frname[50];

	rmnum=2;
	file=open("quickroom",O_RDWR);
	for (a=0; a<MAXROOMS; ++a) {
		read(file,&quickroom,sizeof(struct quickroom));
		if (!strucmp(quickroom.QRname,room)) rmnum=a;
		}
	lseek(file,(long)(rmnum*sizeof(struct quickroom)),0);
	read(file,&quickroom,sizeof(struct quickroom));
	quickroom.QRhighest=where->smnumber;
	lseek(file,(long)(rmnum*sizeof(struct quickroom)),0);
	write(file,&quickroom,sizeof(struct quickroom));
	close(file);
	sprintf(frname,"./rooms/fullrm%d",rmnum);
	file=open(frname,O_RDWR);
	read(file,&fullroom,sizeof(struct fullroom));
	for (a=0; a<(MSGSPERRM-1); ++a) {
		fullroom.FRnum[a]=fullroom.FRnum[a+1];
		fullroom.FRpos[a]=fullroom.FRpos[a+1];
		}
	fullroom.FRnum[MSGSPERRM-1]=where->smnumber;
	fullroom.FRpos[MSGSPERRM-1]=where->smpos;
	lseek(file,0L,0);
	write(file,&fullroom,sizeof(struct fullroom));
	close(file);
	return(0);
}






checkpath(path,sys)	/* Checks to see whether its ok to send */
char path[];		/* Returns 1 for ok, send message	*/
char sys[]; {		/* Returns 0 if message already there	*/
	int a;
	char sys2[30];
	strcpy(sys2,sys);
	strcat(sys2,"!");

	for (a=0; a<strlen(path); ++a) 
		if (!strncmp(&path[a],sys2,strlen(sys2))) return(0);
	return(1);
}

sysinfo(name)	/* determine routing from sysinfo file */
char name[]; {
	int file,a;
	char aaa[100];
	file=open("network/mail.sysinfo",O_RDONLY);
GETSN:	do {
		a=getstring(file,aaa);
		} while ((a>=0)&&(strucmp(aaa,name)));
	a=getstring(file,aaa);
	if (!strncmp(aaa,"use ",4)) {
		strcpy(name,&aaa[4]);
		goto GETSN;
		}
	close(file);
	if (!strncmp(aaa,"bin",3)) {
		return(0);
		}
	printf("Error in network/mail.sysinfo file\n");
	return(-1);
	}

getstring(file,string)		/* get a line of text from a file */
int file; char string[]; {	/* ignores lines beginning with # */
int a,b,c;
do {
	strcpy(string,"");
	a=0;
	do {
		c=0; b=read(file,&c,1);
		if (b<1) { string[a]=0; return(-1); }
		string[a]=c;
		++a;
		} while(c!=10);
		string[a-1]=0;
	} while(string[0]=='#');
return(strlen(string));
}

